
function Cavform(options) {
    
    //
    // Initial state
    //

    var pointtab = [];
    var arctab = [];
    var npoint = 0;
    var npointdim = 0;
    var narc = 0;
    var nunvisited = 0;
    var ndisconnected = 0;
    var npossiblytootight = 0;
    var ndiscounted = 0;
    var listunvisited = [];
    var listdisconnected = [];
    var listpossiblytootight = [];
    var listdiscounted = [];
    var nentrance = 0;
    var maxoklength = 10.0;
    var lineno = 0;
    var debugvar = false;
    
    //
    // Internal Functions
    //

    function owndebug(string) {
	if (debugvar) debug(string);
    }
    
    function assert(condition,name) { 
	if (!condition) throw ("Assertion " + name + " failed");
    }

    function initialise() {
	if (options.debug !== undefined) debugvar = options.debug;
	if (options.maxoklength !== undefined) maxoklength = options.maxoklength;
	assert(isvalidnumber(2),"isvalidnumber 1");
	assert(isvalidnumber(2.7),"isvalidnumber 2");
	assert(!isvalidnumber(Number("ei")),"isvalidnumber 3");
    }

    function fatal(s) {
	throw ("cavfom2survex.sh: error on line " + lineno + ": " + s);
    }

    function warn(s) {
	debug("cavfom2survex.sh: warning on line " + lineno + ": " + s);
    }

    function arrayPush(a,i) {
	a[a.length] = i;
    }

    function checkispointname(point) {
	if (point.indexOf(",") != -1) {
	    fatal("point " + point + " name includes a comma, which is not allowed");
	}
	if (point.indexOf("-") != -1) {
	    fatal("point " + point + " name includes a dash, which is not allowed");
	}
    }

    function checkhasnolocation(point) {
	if (findpoint(point).lat !== undefined) {
	    fatal("location for " + point + " has already been defined");
	}
    }

    function checkisarc(points) {
	var splitr = points.split("-");
	if (splitr.length != 2) {
	    fatal(points + " is not a valid point arc");
	}
	checkispoint(splitr[0]);
	checkispoint(splitr[1]);
	if (findarc(points) === undefined) {
	    fatal(points + " is not a defined arc");
	}
    }

    function addtolist(l,item) {
	arrayPush(l,item);
    }

    function alwaystrue(x) {
	return(true);
    }
    
    function somearcsource_leadingto(point) {
	var result = undefined;
	maparcs(
	    alwaystrue,
	    function (arcobj) {
		if (arcobj.point2 == point && result === undefined) {
		    result = arcobj.point1;
		}
	    });
	return(result);
    }
    
    function dummyinitialisationofpointdimensions(pointobj) {
	pointobj.lookingfrom = somearcsource_leadingto(pointobj.name);
	pointobj.dimension.type = "END";
	if (pointobj.lookingfrom === undefined) {
	    pointobj.dimension.left = 0.0;
	    pointobj.dimension.right = 0.0;
	    pointobj.dimension.up = 0.0;
	    pointobj.dimension.down = 0.0;
	} else {
	    pointobj.dimension.left = 0.2;
	    pointobj.dimension.right = 0.2;
	    pointobj.dimension.up = 0.2;
	    pointobj.dimension.down = 0.2;
	}
    }

    function verifycompleteness() {
	if (npoint == 0) fatal("no points defined");
	if (narc == 0) fatal("no arcs defined");
    }

    function verifypointdimensions() {
	mappoints(
	    alwaystrue,
	    function (pointobj) {
		if (pointobj.dimension.type === undefined) {
		    warn("missing LRUD dimensions for point " + pointobj.name);
		    dummyinitialisationofpointdimensions(pointobj);
		}
	    });
    }

    function stringStartsWith(s,b) {
	var x = s.slice(0,b.length);
	return(x == b);
    }

    function isvalidnumber(n) {
	
	//
	// Here we want to test for NaN. This is needlessly tricky!
	//
	// NaN is the only JavaScript object not equal to itself
	// Technique taken from http://stackoverflow.com/questions/2652319/how-do-you-check-that-a-number-is-nan-in-javascript
	//
	
	return((typeof n === "number") &&
	       n === n);
    }

    function getnumberfromdirection(point1,field) {
	var pointobj = findpoint(point1);
	assert(pointobj !== undefined,"getnumberfromdirection");
	assert(isdirection(field),"getnumberfromdirection");
	var fieldvalue = pointobj.dimension.field;
	if (fieldvalue === undefined) {
	    fatal("DISCOUNT definition from point " + point1 + " refers to dimensions (" + field + ") that have not been set");
	}
	return(fieldvalue);
    }

    function isdirection(value) {
	return(value === "left" ||
	       value === "right" ||
	       value === "up" ||
	       value === "down");
    }

    function getnumberordirection(point1,value) {
	if (isdirection(value)) {
	    return(value);
        } else {
	    return(Number(value));
        }
    }
    
    function processarclineopt(arcobj,opt) {

	if (opt === "UNVISITED") {
	    
	    arcobj.visited = false;
	    nunvisited++;
	    addtolist(listunvisited,arcobj.name);

	} else if (opt === "DISCONNECTEDPART") {
	    
	    arcobj.disconnectedpart = true;
	    findpoint(arcobj.point2).disconnectedpart = true;
	    ndisconnected++;
	    addtolist(listdisconnected,arcobj.name);
	    
	} else if (opt === "POSSIBLYTOOTIGHT") {
	    
	    arcobj.possiblytootight = true;
	    npossiblytootight++;
	    addtolist(listpossiblytootight,arcobj.name);

	} else if (opt === "DISCOUNT") {
	    
	    arcobj.discount = true;
	    arcobj.discountby = arcobj.len;
	    ndiscounted++;
	    addtolist(listdiscounted,arcobj.name);
	    
	} else if (stringStartsWith(opt,"DISCOUNT=")) {
	    
	    var numberpart = opt.slice(opt.indexOf("=")+1);
	    var amount = getnumberordirection(arcobj.point1,numberpart);
	    if (!isdirection(amount) && !isvalidnumber(amount))
		fatal("DISCOUNT option has invalid number " + numberpart + " in ARC " + arcobj.name);
	    if (!isdirection(amount) && amount > arcobj.len)
		fatal("DISCOUNT option has too big number " + numberpart + " (more than " + arcobj.len.toFixed(1) + ") in ARC " + arcobj.name);
	    arcobj.discount = true;
	    arcobj.discountby = amount;
	    ndiscounted++;
	    addtolist(listdiscounted,arcobj.name);
	    
	} else {
	    
            fatal("syntax error in option " + opt + " for ARC " + arcobj.name);
	    
	}

    }

    function processarclineopts(arcobj,opts) {
	opts = opts.trim();
	var components = opts.split(",");
	var i;
	for (i = 0; i < components.length; i++) {
	    processarclineopt(arcobj,components[i]);
	}
    }
	
    function processline(line) {

	if (line.indexOf("#") > 0) {
	    line = line.slice(0,line.indexOf("#"));
	}
        line = line.trim();
	var components = line.split("	");
	
	if (line == "") {

	    // nop

	} else if (stringStartsWith(line,"#")) {

	    // nop
	    
	} else if (stringStartsWith(line,"POINT	")) {

	    var point = components[1];
	    var explanation = components[2];
	    checkispointname(point);
	    var pointobj = findpoint(point);
	    if (pointobj !== undefined) {
		fatal(point + " is already a defined point");
	    }
	    pointtab.push({
		name: point,
		explanation: explanation,
		entrance: false,
		disconnectedpart: false,
		lat: undefined,
		lon: undefined,
		alt: undefined,
		dimension: {
		    type: undefined,
		    left: undefined,
		    right: undefined,
		    up: undefined,
		    down: undefined
		}
	    });
	    npoint++;
	    
	} else if (stringStartsWith(line,"ENTRANCE	")) {

	    var point = components[1];
	    checkispoint(point);
	    findpoint(point).entrance = true;
	    nentrance++;

	} else if (stringStartsWith(line,"POINTLOC	")) {

	    var point = components[1];
	    var lat = components[2];
	    var lon = components[3];
	    var alt = Number(components[4]);
	    checkispoint(point);
	    checkhasnolocation(point);
	    if (!isvalidnumber(alt)) fatal("altitude " + components[4] + " is not a valid number");
	    var pointobj = findpoint(point);
	    pointobj.lat = lat;
	    pointobj.lon = lon;
	    pointobj.alt = alt;

	} else if (stringStartsWith(line,"POINTDIM	")) {

	    var point = components[1];
	    var lookingfrom = components[2];
	    if (lookingfrom === undefined) fatal("missing looking from field");
	    var left = components[3];
	    if (left === undefined) fatal("missing left field");
	    if (left != "END") {
		var right = components[4];
		if (right === undefined) fatal("missing right field");
		var up = components[5];
		if (up === undefined) fatal("missing up field");
		var down = components[6];
		if (down === undefined) fatal("missing down field");
	    }
	    checkispoint(point);
	    if (left != "END" && (left.indexOf(" ") > 0 ||
				  right.indexOf(" ") > 0 ||
				  up.indexOf(" ") > 0 ||
				  down.indexOf(" ") > 0)) {
		fatal("point " + point + " dimension definition has extra spaces");
	    }
	    if (lookingfrom === "-") lookingfrom = undefined;
	    if (lookingfrom !== undefined) {
		checkispoint(lookingfrom);
		points = lookingfrom + "-" + point;
		pointsrev = point + "-" + lookingfrom;
		if (findarc(points) === undefined &&
		    findarc(pointsrev) === undefined) {
		    fatal(points + " is not a defined arc");
		}
		
	    }
	    var pointobj = findpoint(point);
	    pointobj.dimension.lookingfrom = lookingfrom;
	    if (left == "END") {
		pointobj.dimension.type = "END";
		pointobj.dimension.left = 0;
		pointobj.dimension.right = 0;
		pointobj.dimension.up = 0;
		pointobj.dimension.down = 0;
	    } else {
		pointobj.dimension.type = "DIM";
		pointobj.dimension.left = Number(left);
		pointobj.dimension.right = Number(right);
		pointobj.dimension.up = Number(up);
		pointobj.dimension.down = Number(down);
		npointdim++;
	    }

	    if (!isvalidnumber(pointobj.dimension.left)) fatal(components[3] + " is not a valid number");
	    if (!isvalidnumber(pointobj.dimension.right)) fatal(components[4] + " is not a valid number");
	    if (!isvalidnumber(pointobj.dimension.up)) fatal(components[5] + " is not a valid number");
	    if (!isvalidnumber(pointobj.dimension.down)) fatal(components[6] + " is not a valid number");
	    
	} else if (stringStartsWith(line,"ARC	")) {
	    
	    var points = components[1];
	    var pointsarr = points.split("-");
	    if (pointsarr.length != 2) {
		fatal(points + " is not a valid point arc");
	    }
	    var point1 = pointsarr[0];
	    var point2 = pointsarr[1];
	    checkispoint(point1);
	    checkispoint(point2);
	    if (findarc(points) !== undefined) {
		fatal(points + " is already a defined arc");
	    }
	    var len = Number(components[2]);
	    if (!isvalidnumber(len)) fatal("segment length " + components[2] + " is not a valid number");
	    if (len == 0.0 || len >= maxoklength) {
		fatal("arc " + point2 + " length " + len + " seems odd");
	    }
	    var direction = Number(components[3]);
	    if (!isvalidnumber(direction)) fatal("compass direction " + components[3] + " is not a valid number");
	    if (direction < 0.0 || direction >= 360.0) {
		fatal("arc " + point2 + " direction " + direction + " seems odd");
	    }
	    var upangle = Number(components[4]);
	    if (!isvalidnumber(upangle)) fatal("upangle " + components[4] + " is not a valid number");
	    if (upangle < -90.0 || upangle > 90.0) {
		fatal("arc " + point2 + " upangle " + upangle + " seems odd");
	    }
	    var arcobj = {
		name: points,
		point1: point1,
		point2: point2,
		len: len,
		direction: direction,
		upangle: upangle,
		visited: true,
		disconnectedpart: false,
		tootight: false,
		discount: false
	    };
	    arctab.push(arcobj);
	    narc++;

	    var opts = components[5];
	    if (opts !== undefined) {
		processarclineopts(arcobj,opts);
	    }
	    
	} else {
	    
            fatal("syntax error: " + line);
	    
	}

    }
    
    function readinput() {
        var emptyctr = 0;
	var lastnonemptylineno = 1;
	
        while (true) {
	    var line = readline();
	    lineno++;
	    if (line === "") {
		emptyctr++;
	    } else {
		emptyctr = 0;
		lastnonemptylineno = lineno;
	    }
	    if (emptyctr > 100) {
		lineno = lastnonemptylineno;
		return;
	    }
	    processline(line);
	}
	
    }

    function lensumbase(f) {
	var result = 0.0;
	owndebug("lensumbase");
	maparcs(f,
		function (arcobj) { result += arcobj.len; } );
        return(result);
    }
    
    function lensum(f) {
	return(lensumbase(function (arcobj) { return(f(arcobj)); }));
    }
    
    function lensumunvisited(f) {
	return(lensumbase(function (arcobj) { return(f(arcobj) && !arcobj.visited); }));
    }
    
    function lensumdisconnected(f) {
	return(lensumbase(function (arcobj) { return(f(arcobj) && arcobj.disconnectedpart); }));
    }
    
    function lensumpossiblytootight(f) {
	return(lensumbase(function (arcobj) { return(f(arcobj) && arcobj.possiblytootight); }));
    }
    
    function lensumdiscounted(f) {
	var result = 0.0;
	maparcs(f,
		function (arcobj) { if (arcobj.discount) result += arcobj.discountby; } );
        return(result);
    }
    
    function lensumunvisiteddisconnectedpossiblytootightordiscounted(f) {
	var result = 0.0;
	maparcs(f,
		function (arcobj) {
		    if (!arcobj.visited || arcobj.disconnectedpart || arcobj.possiblytootight)
			result += arcobj.len;
		    if (arcobj.discount)
			result += arcobj.discountby;
		} );
        return(result);
    }

    function filterlist(f,l) {
	var result = [];
	var i;
	for (i = 0; i < l.length; i++) {
	    if (f(l[i])) {
		result.push(l[i]);
	    }
	}
	return(result);
    }

    //
    // Public functions
    //

    function getstats(f) {
	var result = "";

	function fname(n) {
	    return(f(findarc(n)));
	}
	
	owndebug("getstats block 1");
	var l_lensum = lensum(f);
	owndebug("getstats block 1 - step 2");
	var l_lensumunvisited = lensumunvisited(f);
	owndebug("getstats block 1 - step 3");
	var l_lensumvisited = l_lensum - l_lensumunvisited;
	owndebug("getstats block 1 - step 4");
	var l_lensumdisconnected = lensumdisconnected(f);
	owndebug("getstats block 1 - step 5");
	var l_lensumpossiblytootight = lensumpossiblytootight(f);
	owndebug("getstats block 1 - step 6");
	var l_lensumregular = l_lensum - l_lensumpossiblytootight;
	owndebug("getstats block 1 - step 7");
	var l_lensumdiscounted = lensumdiscounted(f);
	owndebug("getstats block 1 - step 8");
	var l_lensumunvisiteddisconnectedpossiblytootightordiscounted = lensumunvisiteddisconnectedpossiblytootightordiscounted(f);
	owndebug("getstats block 1 - step 9");
	var l_lensumnormal = l_lensum - l_lensumunvisiteddisconnectedpossiblytootightordiscounted;

	owndebug("getstats block 2");
	var s_lensum = l_lensum.toFixed(1).toString();
	var s_lensumunvisited = l_lensumunvisited.toFixed(1).toString();
	var s_lensumvisited = l_lensumvisited.toFixed(1).toString();
	var s_lensumdisconnected = l_lensumdisconnected.toFixed(1).toString();
	var s_lensumpossiblytootight = l_lensumpossiblytootight.toFixed(1).toString();
	var s_lensumregular = l_lensumregular.toFixed(1).toString();
	var s_lensumdiscounted = l_lensumdiscounted.toFixed(1).toString();
	var s_lensumunvisiteddisconnectedpossiblytootightordiscounted = l_lensumunvisiteddisconnectedpossiblytootightordiscounted.toFixed(1).toString();
	var s_lensumnormal = l_lensumnormal.toFixed(1).toString();
	
	owndebug("getstats block 3");
	result += "total length as calculated by cavform2survex.sh: " + s_lensum + "m\n";
	result += "  of which visited: " + s_lensumvisited + "m, unvisited: " + s_lensumunvisited + "m";
	if (listunvisited.length > 0) result += " (" + filterlist(fname,listunvisited).toString() + ")";
	result += "\n";
	result += "  of which disconnected parts: " + s_lensumdisconnected;
	if (listdisconnected.length > 0) result += " (" + filterlist(fname,listdisconnected).toString() + ")";
	result += "\n";
	result += "  of which regular passages: " + s_lensumregular + "m and possibly too tight: " + s_lensumpossiblytootight + "m";
	if (listpossiblytootight.length > 0) result += " (" + filterlist(fname,listpossiblytootight).toString() + ")";
	result += "\n";
	result += "  of which discounted passages: " + s_lensumdiscounted + "m";
	if (listdiscounted.length > 0) result += " (" + filterlist(fname,listdiscounted).toString() + ")";
	result += "\n";
	result += "  of which regular passages " + s_lensumnormal + "m and any unvisited, possibly too tight, disconnected, or discounted parts: " + s_lensumunvisiteddisconnectedpossiblytootightordiscounted + "m\n";
	result += npoint + " measurement points\n";
	result += nentrance + " of those points are entrances\n";
	result += npointdim + " of those points have left-right-up-down dimensions, " + (npoint - npointdim) + " without\n";
	result += narc + " point-to-point measurements\n";
	return(result);
    }
    
    function findpoint(point) {
	var i;
	for (i = 0; i < pointtab.length; i++) {
	    if (pointtab[i].name == point) {
		return(pointtab[i]);
	    }
	}
	return(undefined);
    }
    
    function findarc(points) {
	var i;
	for (i = 0; i < arctab.length; i++) {
	    if (arctab[i].name == points) {
		return(arctab[i]);
	    }
	}
	return(undefined);
    }
    
    function findarc2(point1,point2) {
	var i;
	for (i = 0; i < arctab.length; i++) {
	    if (arctab[i].point1 == point1 &&
	        arctab[i].point2 == point2) {
		return(arctab[i]);
	    }
	}
    }
    
    function mappoints(ff,ef) {
	var i;
	for (i = 0; i < pointtab.length; i++) {
	    if (ff(pointtab[i])) {
		ef(pointtab[i]);
	    }
	}
    }
    
    function maparcs(ff,ef) {
	var i;
	owndebug("maparcs");
	for (i = 0; i < arctab.length; i++) {
	    owndebug("maparcs item " + i.toString());
	    if (ff(arctab[i])) {
		ef(arctab[i]);
	    }
	}
    }
    
    function findsuccessors(pf,af,point) {
	var list = [];
	mappoints(pf,
		  function (apointobj) {
		      if (apointobj.dimension.lookingfrom == point) {
			  var arcobj = findarc2(point,apointobj.name);
			  if (arcobj !== undefined &&
			      af(arcobj)) {
			      arrayPush(list,apointobj.name);
			  }
		      }
		  });
	owndebug("findsuccessors " + point + " = " + list.toString());
	return(list);
    }
    
    function checkispoint(point) {
	checkispointname(point);
	if (findpoint(point) === undefined) {
	    fatal(point + " is not a defined point\n");
	}
    }

    function calculatediscounts() {
	owndebug("calculatediscounts");
	maparcs(alwaystrue,
		function (arcobj) {
		    var dir = arcobj.discountby;
		    if (arcobj.discounted && isdirection(dir)) {
			arcobj.discountby = getnumberfromdirection(arcobj.point1,dir);
			owndebug("setting arc " + arcobj.name + " discount based on " + dir + " to " + arcobj.discountby.toFixed(1).toString());
		    }
		});
    }
    
    function reversedirection(direction) {
	if (direction < 180.0) return(direction + 180.0);
	else return(direction - 180.0);
    }

    function reverseinclination(upangle) {
	return(-upangle);
    }

    function findunusedpointnumbers() {
	var nonnumberpoint = false;
	var lowestnum = undefined;
	var highestnum = undefined;
	var reg = new RegExp('^[0-9]+$');
	mappoints(alwaystrue,
		  function (pointobj) {
		      if (!reg.test(pointobj.name)) {
			  nonnumberpoint = true;
		      } else {
			  var asnum = parseInt(pointobj.name,10);
			  if (lowestnum === undefined) {
			      lowestnum = asnum;
			  } else if (asnum < lowestnum) {
			      lowestnum = asnum;
			  }
			  if (highestnum === undefined) {
			      highestnum = asnum;
			  } else if (asnum > highestnum) {
			      highestnum = asnum;
			  }
		      }
		  });
	if (nonnumberpoint === false && lowestnum !== undefined && highestnum !== undefined) {
	    var i;
	    var unused = [];
	    for (i = lowestnum; i <= highestnum; i++) {
		var name = i.toString();
		if (findpoint(name) === undefined) {
		    unused.push(i);
		}
	    }
	    if (unused.length > 0) {
		warn("Point numbers run from " + lowestnum.toString() +
		     " to " + highestnum.toString() +
		     " but numbers " + unused.toString() +
		     " are unused");
	    }
	}
    }

    //
    // Code to run at creation time
    //
    
    owndebug("cavform initialise");
    initialise();
    readinput();
    calculatediscounts();
    verifycompleteness();
    verifypointdimensions();
    findunusedpointnumbers();
    owndebug("cavform done");

    //
    // What we return as an object
    //
    
    return({
	getstats: getstats,
	findpoint: findpoint,
	findarc: findarc,
	findarc2: findarc2,
	mappoints: mappoints,
	maparcs: maparcs,
	findsuccessors: findsuccessors,
	checkispoint: checkispoint,
	alwaystrue: alwaystrue,
	reversedirection: reversedirection,
	reverseinclination: reverseinclination
           });
}

