function Cavform2Survex(options,cf) {
    
    //
    // Initial state
    //

    var gps = false;
    var includeunvisited = true;
    var includedisconnected = true;
    var includelrud = true;
    var includepossiblytootight = true;
    var debugvar = false;
    var checkpath = [];
    var checkpatharcs = [];
    var explored = [];
    var exclude = [];

    //
    // Internal Functions
    //

    function owndebug(string) {
	if (debugvar) debug(string);
    }
    
    function fatal(s) {
	throw ("cavfom2survex.sh: error: " + s);
    }

    function initialise() {
	if (options.debug !== undefined) debugvar = options.debug;
	owndebug("cavform2survex initialise");
	if (options.gps !== undefined) gps = options.gps;
	owndebug("gps = " + gps + " and options.gps = " + options.gps);
	if (options.includeunvisited !== undefined) includeunvisited = options.includeunvisited;
	if (options.includedisconnected !== undefined) includedisconnected = options.includedisconnected;
	owndebug("includedisconnected is " + includedisconnected);
	if (options.includelrud !== undefined) includelrud = options.includelrud;
	if (options.includepossiblytootight !== undefined) includepossiblytootight = options.includepossiblytootight;
	if (options.includeonlypath !== undefined && options.includeonlypath != "all") {
	    checkpath = options.includeonlypath.split(",");
	}
	if (options.exclude !== undefined && options.exclude != "none") {
	    exclude = options.exclude.split(",");
	}
    }

    function tosurvexcoord(x) {
	x = x.replace("N ","");
	x = x.replace("S ","");
	x = x.replace("W ","");
	x = x.replace("E ","");
	return(x);
    }
    
    function assert(condition,name) { 
	if (!condition) throw ("Assertion " + name + " failed");
    }

    function listfind(list,f) {
	var i;
	for (i = 0; i < list.length; i++) {
	    if (f(list[i])) return(list[i]);
	}
	return(undefined);
    }
	
    function findexplored(point) {
	return(listfind(explored,function (item) { return(item.point === point); }));
    }

    function isexplored(point) {
	assert(point !== undefined,"isexplored undefined test");
	return(findexplored(point) !== undefined);
    }
    
    function markexplored(point) {
	assert(point !== undefined,"markexplored undefined test");
	assert(!isexplored(point),"markexplored isexplored test 1");
	explored.push({
	    point: point,
	    count: 1
	});
	assert(isexplored(point),"markexplored isexplored test 2");
    }
    
    function incrementexplored(point) {
	assert(point !== undefined,"incrementexplored undefined test");
	assert(isexplored(point),"incrementexplored condition");
	var entry = findexplored(point);
	assert(entry !== undefined,"incrementexplored entry");
	assert(entry.count > 0,"incrementexplored count");
	entry.count++;
    }
    
    function findunexploredsuccessors(point) {
	var list = cf.findsuccessors(pointshouldbeincludedobj,
				     arcshouldbeincludedobj,
				     point);
	var newlist = [];
	var i;
	
	for (i = 0; i < list.length; i++) {
	    var nextpoint = list[i];
	    assert(nextpoint !== undefined,"findunexploredsuccessors undefined test");
	    owndebug("**** PRINTLRUD SUCCESSOR TESTING...");
	    var condition1 = !isexplored(nextpoint);
	    var condition2 = arcshouldbeincluded(point,nextpoint);
	    owndebug("**** PRINTLRUD SUCCESSOR " + nextpoint + " !isexplored " +
		     condition1.toString() +
		     " arcshouldbeincluded " +
		     condition2.toString());
	    if (condition1 && condition2) {
		newlist.push(nextpoint);
		owndebug("added " + nextpoint + " to list");
	    }
	}
	
	return(newlist);
    }
 
    function output(string) {
	print(string);
    }

    function lentostring(len) {
	return(len.toFixed(2).toString());
    }

    function alttostring(alt) {
	return(alt.toFixed(2).toString());
    }

    function directiontostring(direction) {
	return(direction.toFixed(1).toString());
    }

    function angletostring(angle) {
	return(angle.toFixed(1).toString());
    }

    function explorepassageone(point) {
	assert(point !== undefined,"explorepassageone undefined test");
	if (!isexplored(point)) {
	    markexplored(point);
	} else {
	    incrementexplored(point);
	}
	var obj = cf.findpoint(point);
	assert(obj !== undefined,"explorepassageone");
	output(obj.name + " " +
	       lentostring(obj.dimension.left) + " " +
	       lentostring(obj.dimension.right) + " " +
	       lentostring(obj.dimension.up) + " " +
	       lentostring(obj.dimension.down) + " ; " +
	       obj.explanation);
    }
	
    function itemsinlist(l) {
	return(l.length);
    }

    function fullyexplored(point) {
	assert(point !== undefined,"fullyexplored undefined test");
	if (!isexplored(point)) return(false);
	var entry = findexplored(point);
	if (entry.count <
	    itemsinlist(cf.findsuccessors(pointshouldbeincludedobj,arcshouldbeincludedobj,point))) return(false);
	return(true);
    }

    function explorepassage(point) {
	assert(point !== undefined,"explorepassage undefined test 1");
	if (!fullyexplored(point)) {
	    explorepassageone(point);
	    var successors = findunexploredsuccessors(point);
	    owndebug("**** PRINTLRUD SUCCESSORS FROM " + point + " are " + successors.toString());
	    while (successors.length > 0) {
		point = successors[0];
		assert(point !== undefined,"explorepassage undefined test 2");
		explorepassageone(point);
		successors = findunexploredsuccessors(point);
	    }
	}
    }
    
    function printentrances() {
	cf.mappoints(pointshouldbeincludedobj,
		     function (pointobj) {
			 if (pointobj.entrance) {
			     output("*entrance	" + pointobj.name + " ; " + pointobj.explanation);
			 }
		     });
    }
    
    function startlrudpath(point) {
	output("*begin ; for the passage starting from " + point);
	output("*data passage station left right up down");
	explorepassage(point);
	output("*end");
    }
    
    function printlrud() {

	owndebug("**** PRINTLRUD");
	
	//
	// Pass 1: best starting nodes
	// 

	cf.mappoints(pointshouldbeincludedobj,
		     function (pointobj) {
			 if (pointobj.dimension.type === "DIM" &&
			     pointobj.dimension.lookingfrom === undefined) {
			     owndebug("**** PRINTLRUD START " + pointobj.name);
			     startlrudpath(pointobj.name);
			 }
		     });
	
	//
	// Pass 2: remaining nodes
	// 
	
	owndebug("**** PRINTLRUD STAGE 1");
	cf.mappoints(pointshouldbeincludedobj,
		     function (pointobj) {
			 while (pointobj.dimension.type === "DIM" &&
				!fullyexplored(pointobj.name)) {
			     owndebug("**** PRINTLRUD START " + pointobj.name);
			     startlrudpath(pointobj.name);
			 }
		     });
	
	owndebug("**** END PRINTLRUD");
    }

    function arcshouldbeincludedinternal(arcobj) {
	assert(arcobj.name !== undefined,"arcshouldbeincludedinternal 1");

	assert(arcobj.disconnectedpart === false || arcobj.disconnectedpart === true,"arcshouldbeincludedinternal 2");
	var condition1 = incheckpatharc(arcobj.name);
	var condition2 = (includedisconnected || !arcobj.disconnectedpart);
	var condition3 = (includeunvisited || arcobj.visited);
	owndebug("arcshouldbeincludedinternal " + arcobj.name +
		 " incheckpatharc " + condition1.toString() +
		 " disconnected " + includedisconnected.toString() + " " + arcobj.disconnectedpart.toString() + "=" + condition2.toString() +
		 " visited " + includeunvisited.toString() + " " + arcobj.visited.toString() + "=" + condition3.toString());
	return(condition1 && condition2 && condition3);
    }
    
    function cavename() {
	return("k");
    }
    
    function printsurvex() {
	owndebug("cavform2survex printsurvex");
	assert(cf !== undefined);
	output("; generated by cavform2survex");
        output("");
	output("*begin " + cavename());
	output("*title Korkberget");
	var now = new Date();
	output("*date " + now.getFullYear().toString() + "." + (now.getMonth()+1).toString() + "." + now.getDate().toString());
	output("*data normal from to tape compass clino");
	output("; from to tape compass clino\n");
	owndebug("**** PRINTARCS");
	cf.maparcs(arcshouldbeincludedobj,
		   function (arcobj) {
		       owndebug("arcobj.point1 = " + arcobj.point1);
		       owndebug("arcobj.point2 = " + arcobj.point2);
		       owndebug("arcobj.len = " + arcobj.len);
		       owndebug("arcobj.direction = " + arcobj.direction);
		       owndebug("arcobj.upangle = " + arcobj.upangle);
		       owndebug("arcobj.disconnectedpart = " + arcobj.disconnectedpart);
		       output(arcobj.point1 + "	" +
			      arcobj.point2 + "	" +
			      lentostring(arcobj.len) + "	" +
			      directiontostring(arcobj.direction) + "	" +
			      angletostring(arcobj.upangle));
		   });
	owndebug("**** END PRINTARCS");
	
	if (gps) {
	    owndebug("**** PRINTGPS");
	    output("*cs LONG-LAT");
	    output("*cs out S-MERC");
	    cf.mappoints(
		function (pointobj) {
		    return(pointshouldbeincludedobj(pointobj) &&
			   pointobj.lat !== undefined);
		},
		function (pointobj) {
		    output("*fix " + pointobj.name + " " +
			   tosurvexcoord(pointobj.lat) + " " +
			   tosurvexcoord(pointobj.lon) + " " +
			   alttostring(pointobj.alt));
		});
	    owndebug("**** END PRINTGPS");
	}
	printentrances();  
	if (includelrud) {
	    printlrud();
	}
	output("*end K");
    }
    
    function printgraphproperties() {
	owndebug("**** PRINTGRAPHPROPERTIES");
	cf.maparcs(arcshouldbeincludedobj,
		   function (arcobj) {
		       if (arcobj.discount) {
			   owndebug("testing arc " + arcobj.point1 + "-" + arcobj.point2 +
				    " DC=" + arcobj.discount.toString() +
				    " DBY=" + arcobj.discountby.toString());
			   output("graphproperty:A:" +
                                  cavename() + "." + arcobj.point1 + "-" +
				  cavename() + "." + arcobj.point2 + ":D=" + (arcobj.discountby/arcobj.len).toFixed(2));
		       }
		   });
	if (includelrud) {
	    owndebug("**** PRINTGRAPHPROPERTIES STAGE 2");
	    cf.mappoints(pointshouldbeincludedobj,
			 function (pointobj) {
			     if (pointobj.dimension.type == "DIM") {
				 var angle = undefined;
				 var inclination = undefined;
				 if (pointobj.dimension.lookingfrom === undefined) {
				     cf.maparcs(arcshouldbeincludedobj,
						function (arcobj) {
						    owndebug("trying to find angle for " + pointobj.name + " from " + arcobj.point1);
						    if (angle === undefined && arcobj.point1 === pointobj.name) {
							owndebug("found direction " + arcobj.direction.toFixed(1));
							angle = arcobj.direction;
							inclination = arcobj.upangle;
						    }
						});
				     cf.maparcs(arcshouldbeincludedobj,
						function (arcobj) {
						    owndebug("trying to find angle for " + pointobj.name + " from " + arcobj.point1);
						    if (angle === undefined && arcobj.point2 === pointobj.name) {
							owndebug("found direction " + arcobj.direction.toFixed(1));
							angle = cf.reversedirection(arcobj.direction);
							inclination = cf.reverseinclination(arcobj.upangle);
						    }
						});
				     if (angle === undefined) {
					 fatal("Cannot find direction for LRUD dimensions for point " + pointobj.name);
				     }
				     if (inclination === undefined) {
					 fatal("Cannot find upangle for LRUD dimensions for point " + pointobj.name);
				     }
				 } else {
				     var arcobj = cf.findarc2(pointobj.dimension.lookingfrom,pointobj.name);
				     if (arcobj !== undefined) {
					 angle = arcobj.direction;
					 inclination = arcobj.upangle;
				     } else {
					 arcobj = cf.findarc2(pointobj.name,pointobj.dimension.lookingfrom);
					 if (arcobj !== undefined) {
					     angle = cf.reversedirection(arcobj.direction);
					     inclination = cf.reverseinclination(arcobj.upangle);
					 } else {
					     fatal("Cannot find arc " + pointobj.name + "-" + pointobj.dimension.lookingfrom +
						   " to gather graph properties");
					 }
				     }
				 }
				 owndebug("final direction is " + (angle===undefined?"undef":angle.toFixed(1)));
				 owndebug("and final incl. is " + (inclination===undefined?"undef":inclination.toFixed(1)));
				 output("graphproperty:V:" +
                                        cavename() + "." + pointobj.name +
					":L=" + pointobj.dimension.left.toFixed(2) +
					",R=" + pointobj.dimension.right.toFixed(2) +
					",A=" + angle.toFixed(2) +
					",U=" + pointobj.dimension.right.toFixed(2) +
					",D=" + pointobj.dimension.right.toFixed(2) + 
					",I=" + inclination.toFixed(2));
			     }
			 });
        }
	owndebug("**** END PRINTGRAPHPROPERTIES");
    }

    function verifyarc(point1,point2) {
	var points = point1 + "-" + point2;
	var pointsrev = point2 + "-" + point1;
	if (cf.findarc(points) !== undefined) {
	    return(points);
	} else if (cf.findarc(pointsrev) !== undefined) {
	    return(pointsrev);
	} else {
	    fatal("path segment " + points + " is not a defined path\n");
	}
    }

    function getstats() {
	var result = "";
	
	if (checkpath.length > 0) {
	    var totalcheckpath = 0.0;
	    for (j = 0; j < checkpath.length - 1; j++) {
		totalcheckpath += cf.findarc(checkpatharcs[j]).len;
	    }
	    result += lentostring(totalcheckpath) + "m length of the designated path (" + options.includeonlypath + ")\n";
	}
	
	return(result);
    }
    
    function printstats() {
	owndebug("printstats");
	var text =
	    cf.getstats(arcshouldbeincludedobj) +
	    getstats();
	var lines = text.split("\n");
	var i;
	
	output("");
	for (i = 0; i < lines.length; i++) {
	    if (lines[i] !== "") {
		output("; " + lines[i]);
	    }
	}
	output("");
    }
    
    function verifycheckpath() {
	var i;
	owndebug("cavform2survex verifycheckpath");
	for (i = 0; i < checkpath.length; i++) {
	    cf.checkispoint(checkpath[i]);
	    if (i > 0) {
		var properarc = verifyarc(checkpath[i-1],checkpath[i]);
		checkpatharcs.push(properarc);
	    }
	}
    }
    
    function verifyexclude() {
	var i;
	owndebug("cavform2survex verifyexclude");
	for (i = 0; i < exclude.length; i++) {
	    cf.checkispoint(exclude[i]);
	}
    }
    
    function incheckpatharc(points) {
	var j;
	if (checkpath.length == 0) return(true);
	for (j = 0; j < checkpath.length - 1; j++) {
	    if (points == checkpatharcs[j]) return(true);
	}
	return(false);
    }
    
    function pointshouldbeincluded(point) {
	
	var j;
	
	if (!includedisconnected && cf.findpoint(point).disconnectedpart) {
	    return(false);
	}

	for (j = 1; j <= exclude.length; j++) {
	    if (point == exclude[j]) return(false);
	}
	
	if (checkpath.length == 0) {
	    return(true);
	}
	
	for (j = 1; j <= checkpath.length; j++) {
	    if (point == checkpath[j]) return(true);
	}
	
	return(false);
	
    }
    
    function pointshouldbeincludedobj(pointobj) {
	owndebug("pointshouldbeincludedobj(" + pointobj.name + ")");
	return(pointshouldbeincluded(pointobj.name));
    }
    
    function arcshouldbeincludedobj(arcobj) {
	owndebug("arcshouldbeincludedobj " + arcobj.name);
	return(pointshouldbeincluded(arcobj.point1) &&
	       pointshouldbeincluded(arcobj.point2) &&
	       arcshouldbeincludedinternal(arcobj));
    }
    
    function arcshouldbeincluded(point1,point2) {
	owndebug("arcshouldbeincluded " + point1 + "-" + point2);
	var arcobj = cf.findarc2(point1,point2);
	if (arcobj === undefined) {
	    owndebug("no such arc");
	    return(false);
	}
	return(arcshouldbeincludedobj(arcobj));
    }
    
    //
    // Code to run at creation time
    //
    
    initialise();
    verifycheckpath();    
    verifyexclude();    
    printsurvex();
    printgraphproperties();
    printstats();
    owndebug("cavform2survex done");
    
    //
    // What we return as an object
    //
    
    return({
           });

}
